/*	$NetBSD$	*/

/*
 * Copyright 2001-2005 Wasabi Systems, Inc.
 * All rights reserved.
 *
 * Written by Eduardo Horvath, Simon Burge and Steve C. Woodford
 * for Wasabi Systems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed for the NetBSD Project by
 *      Wasabi Systems, Inc.
 * 4. The name of Wasabi Systems, Inc. may not be used to endorse
 *    or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY WASABI SYSTEMS, INC. ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL WASABI SYSTEMS, INC
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/queue.h>
#include <sys/systm.h>
#include <sys/pool.h>
#include <sys/device.h>

#include <uvm/uvm.h>

#include <machine/cpu.h>
#include <machine/pcb.h>

#include <powerpc/spr.h>
#include <machine/tlb.h>

#ifdef PMAP_DEBUG
#define	DBG_ENTER		0x00000001
#define	DBG_REMOVE		0x00000002
#define	DBG_KENTER		0x00000004
#define	DBG_KREMOVE		0x00000008
#define	DBG_EXTRACT		0x00000010
#define	DBG_PROTECT		0x00000020
#define	DBG_PAGE_PROTECT	0x00000040
#define	DBG_CREATE		0x00000080
#define	DBG_DESTROY		0x00000100
#define	DBG_ACTIVATE		0x00000200
#define	DBG_UPDATE		0x00000400
#define	DBG_CACHE		0x00000800
#define	DBG_GROW_KERNEL		0x00001000
#define	DBG_REMOVE_ALL		0x00002000
#define	DBG_REFMOD		0x00004000
#define	DBG_FIXUP		0x00008000
#define	DBG_PTE_ENTER		0x00010000
#define	DBG_COPY_PAGE		0x00020000
#define	DBG_ZERO_PAGE		0x00040000
#define	DBG_ALL			(~0)
int pmapdebug = 0;
#define	DPRINTF(x,y)	do { if ((x) & pmapdebug) printf y; } while (0)
#else
#define	DPRINTF(x,y)	/* nothing */
#endif

/*
 * This pmap supports the MMU found in some Motorola PowerQuicc processors,
 * for example MPC875 and MPC885.
 *
 * The MMU is based around a software-filled TLB, but has some degree
 * of hardware assisted table-walk which can be used by the low-level
 * TLB miss handlers to speed table lookup.
 *
 * To make use of the hardware assist feature, the page tables must be
 * laid out as described in the PowerQuicc documentation. This pmap
 * attempts to do so.
 *
 * We implement a split address-space instead of the more traditional
 * separate address-space layout for a number of reasons:
 *  - The MMU has no BATs, and an 8MB limit to the size of mapping in
 *    a single TLB entry.
 *  - The exception handlers are invoked with translation disabled.
 * The above two items make a split address-space memory map much
 * easier to deal with on these CPUs.
 *
 * Since the exception handlers are invoked with translation disabled,
 * we can run with no locked TLB entries, thus leaving all TLB entries
 * free for both kernel and userland translations. This is possible
 * because the TLB miss handlers access page tables by their physical
 * addresses thanks to the hardware assist.
 */

/*
 * Pool that pmap structures are allocated from.
 */
static struct pool pmap_pool;

/*
 * Pool and cache of page descriptors.
 */
static struct pool pmap_pt_pool;
static struct pool_cache pmap_pt_cache;

/*
 * Flag to indicate if pmap_init() has done its thing
 */
boolean_t pmap_initialized;

/*
 * Some useful items for the L1 descriptor table
 */
paddr_t pmap_kernel_l1_pa;
#define	PMAP_L1_NULL		0
#define	PMAP_L1_IS_NULL(l1)	((l1) == 0)

/*
 * Address Space ID management
 */
static unsigned int pmap_next_casid;
static unsigned int pmap_casid_generation;

struct pmap kernel_pmap_;

int physmem;
static int npgs;
extern paddr_t msgbuf_paddr;

/*
 * This is a cache of referenced/modified bits.
 */
static char *pmap_attrib;

#define PV_WIRED	0x1
#define PV_WIRE(pv)	((pv)->pv_va |= PV_WIRED)
#define PV_IS_WIRED(pv)	(((pv)->pv_va & PV_WIRED) != 0)
#define PV_UNWIRE(pv)	((pv)->pv_va &= ~PV_WIRED)
#define	PV_VA(pv)	((pv)->pv_va & ~PGOFSET)
#define	PV_CMPVA(va,pv)	(!(((pv)->pv_va^(va))&(~PV_WIRED)))

struct pv_entry {
	struct pv_entry *pv_next;	/* Linked list of mappings */
	vaddr_t pv_va;			/* virtual address of mapping */
	pmap_t pv_pm;
};

struct pv_entry *pv_table;
static struct pool pv_pool;

static inline struct pv_entry *pa_to_pv(paddr_t);
static inline char *pa_to_attr(paddr_t);

static inline void pte_enter(pmap_t, vaddr_t, pt_entry_t);
static inline pt_entry_t *pte_find(pmap_t, vaddr_t);

static inline int pmap_enter_pv(pmap_t, vaddr_t, paddr_t, int);
static void pmap_remove_pv(pmap_t, vaddr_t, paddr_t);

#define	TLBIE(va)	__asm __volatile("tlbie %0" :: "r"(va))
#define	TLBIA()		__asm __volatile("tlbia")
#define	TLBSYNC()	__asm __volatile("tlbsync")

/*
 * pmap_init_tlb()
 *
 * This function is called by the kernel startup code with translation
 * disabled. Its purpose is to fill the TLB with enough 8MB mappings to
 * map the kernel text/data/bss into KSEG0.
 */
void pmap_init_tlb(paddr_t, paddr_t immr_pa)
    __attribute__ ((no_instrument_function));
void
pmap_init_tlb(paddr_t end_kernel, paddr_t immr_pa)
{
	uint32_t epn, rpn;
	int i, ntlb;

	/*
	 * First, initialise the cache
	 */
	mtspr(SPR_IC_CST, IC_CST_CMD_UNLOCKALL);
	mtspr(SPR_IC_CST, IC_CST_CMD_INVALL);
	mtspr(SPR_IC_CST, IC_CST_CMD_ENABLE);

	mtspr(SPR_DC_CST, DC_CST_CMD_UNLOCKALL);
	mtspr(SPR_DC_CST, DC_CST_CMD_INVALL);
	mtspr(SPR_DC_CST, DC_CST_CMD_CLRFWT);
	mtspr(SPR_DC_CST, DC_CST_CMD_ENABLE);

	/*
	 * Now for the MMU...
	 */
	mtspr(SPR_M_CASID, 0);
	mtspr(SPR_MI_AP, MID_AP_ALL_PAGE_PROT);
	mtspr(SPR_MD_AP, MID_AP_ALL_PAGE_PROT);
	mtspr(SPR_MI_CTR, 0);
	mtspr(SPR_MD_CTR, MD_CTR_TWAM_4KB);

	/*
	 * Stuff enough entries in the ITLB and DTLB to cover all of
	 * kernel text/data/bss.
	 * XXX: We should KDASSERT() that ntlb <= 30. But as it's very
	 * unlikely that anyone will be loading a > 260MB kernel...
	 */
	ntlb = ((end_kernel + (8*1024*1024) - 1) / (8*1024*1024)) + 1;
	epn = PPC_KSEG0_BASE | MID_EPN_EV;
	rpn = 0x00000000 | L2_PP_K_RDWR | L2_CHANGE | L2_SZ_LARGE |
	    L2_SHARED | L2_VALID;

	for (i = 0; i < ntlb; i++) {
		mtspr(SPR_MI_CTR, i << MI_CTR_ITLB_INDX_SHIFT);
		mtspr(SPR_MD_CTR, MD_CTR_TWAM_4KB |
		    (i << MD_CTR_DTLB_INDX_SHIFT));
		mtspr(SPR_MI_EPN, epn);
		mtspr(SPR_MD_EPN, epn);
		mtspr(SPR_MI_TWC, MI_TWC_PS_8MB | MI_TWC_V);
		mtspr(SPR_MD_TWC, MD_TWC_W_PS_8MB | MD_TWC_V);
		mtspr(SPR_MI_RPN, rpn);
		mtspr(SPR_MD_RPN, rpn);

		epn += (8*1024*1024);
		rpn += (8*1024*1024);
	}

	/*
	 * Map the IMMR in the next available DTLB slot
	 */
	epn = MPC_IMMR_KVA | MID_EPN_EV;
	rpn = immr_pa | L2_PP_K_RDWR | L2_CHANGE | L2_SZ_LARGE | L2_SHARED |
	    L2_CI | L2_VALID;

	mtspr(SPR_MD_CTR, MD_CTR_TWAM_4KB | (i << MD_CTR_DTLB_INDX_SHIFT));
	mtspr(SPR_MD_EPN, epn);
	mtspr(SPR_MD_TWC, MD_TWC_W_G | MD_TWC_W_PS_8MB | MD_TWC_V);
	mtspr(SPR_MD_RPN, rpn);
}

/*
 * pmap_init_kseg0()
 *
 * This function is invoked from the machine-dependent startup code. At
 * the point where this function is called, translation is enabled using
 * enough 8MB mappings stuffed into the TLB to cover the kernel text/data
 * and bss.
 *
 * The purpose of pmap_init_kseg0() is to set up the initial set of page-
 * tables which map KSEG0 in 8MB chunks. To this end, the caller passes
 * the address of a suitable chunk of memory to be used for those tables.
 * This memory must be aligned to 4KB.
 *
 * In some cases, this memory will be located just after the vector table
 * in SDRAM (the vector table is located at the bottom of KSEG0) and
 * will extend to just before KERNBASE.
 *
 * The pmap requires enough space in this region for the following tables:
 *
 *  - One 4KB page to be used as the kernel's L1 page table.
 *  - Enough 4KB pages to map all of SDRAM into KSEG0. One page is required
 *    for every 8MB of SDRAM.
 *
 * If pt_base is located between the vector table and KERNBASE and we end
 * up running out of space (if we have 512MB SDRAM, for example), then
 * start stealing pages from 'ekern' upwards.
 *
 * Return the KVA of the first unused page.
 */
vaddr_t
pmap_init_kseg0(vaddr_t ekern, vaddr_t pt_base, size_t sdram_size,
    paddr_t immr_pa)
{
	pd_entry_t *l1;
	pt_entry_t *l2, pte;
	vaddr_t va;
	u_int i;

	pt_base = round_page(pt_base);
	if ((KERNBASE - pt_base) == 0)
		pt_base = round_page(ekern);

	/*
	 * Initialise the kernel L1 segment descriptor
	 */
	pmap_kernel()->pm_l1 = l1 = (pd_entry_t *)pt_base;
	pmap_kernel_l1_pa = PPC_KSEG0_TO_PA(pt_base);
	pt_base += PAGE_SIZE;

	/* All entries are invalid to begin with */
	for (i = 0; i < (PAGE_SIZE / sizeof(*l1)); i++)
		l1[i] = PMAP_L1_NULL;

	if ((KERNBASE - pt_base) == 0)
		pt_base = round_page(ekern);

	/*
	 * Map SDRAM linearly into KSEG0 using 8MB pages
	 */
	for (va = PPC_KSEG0_BASE; va < (PPC_KSEG0_BASE + sdram_size);
	    va += (8*1024*1024)) {
		/*
		 * Fix up the two consecutive L1 entries
		 * to point to the next available L2
		 */
		l1[L1_VA2IDX(va)] = l1[L1_VA2IDX(va) + 1] =
		    PPC_KSEG0_TO_PA(pt_base) | L1_PS_8MB |
			    L1_COPY_BACK | L1_VALID;

		/*
		 * Initialise all entries in the L2 descriptor with
		 * identical values.
		 *
		 * XXX: I *think* we could get away with just initialising
		 * the first entry of the L2 since we're using 8MB pages.
		 * At least, that's one way of interpreting the documentation.
		 */
		pte = PPC_KSEG0_TO_PA(va) | L2_PP_K_RDWR | L2_CHANGE |
		    L2_SZ_LARGE | L2_SHARED | L2_VALID;

		l2 = (pt_entry_t *)pt_base;
		pt_base += PAGE_SIZE;
		while (l2 < (pt_entry_t *)pt_base)
			*l2++ = pte;

		if ((KERNBASE - pt_base) == 0)
			pt_base = round_page(ekern);
	}

	/*
	 * Map the IMMR using a single 8MB mapping at MPC_IMMR_KVA
	 * (which will usually be near the top of KSEG1; 0xff000000)
	 */
	l1[L1_VA2IDX(MPC_IMMR_KVA)] = l1[L1_VA2IDX(MPC_IMMR_KVA) + 1] =
	    PPC_KSEG0_TO_PA(pt_base) | L1_G | L1_PS_8MB | L1_VALID;

	pte = immr_pa | L2_PP_K_RDWR | L2_CHANGE | L2_SZ_LARGE | L2_SHARED |
	    L2_CI | L2_VALID;

	l2 = (pt_entry_t *)pt_base;
	pt_base += PAGE_SIZE;
	while (l2 < (pt_entry_t *)pt_base)
		*l2++ = pte;

	if ((KERNBASE - pt_base) == 0)
		pt_base = round_page(ekern);

	/*
	 * Fix up the MMU control registers in preparation for enabling
	 * translation.
	 */
	mtspr(SPR_MI_CTR, 0);
	mtspr(SPR_MD_CTR, MD_CTR_TWAM_4KB);
	mtspr(SPR_MI_AP, MID_AP_ALL_PAGE_PROT);
	mtspr(SPR_MD_AP, MID_AP_ALL_PAGE_PROT);
	mtspr(SPR_M_CASID, 0);
	mtspr(SPR_M_TWB, pmap_kernel_l1_pa);

	return (pt_base);
}

static __inline void
cache_wb_page(pmap_t pm, paddr_t pa)
{
	vaddr_t va;
	int off, ls;

	if (pm && pm->pm_remove_all) {
		DPRINTF(DBG_CACHE, ("cache_wb_page(%p, %08lx): skipping.\n",
		    pm, pa));
		return;
	}

	va = PPC_PA_TO_KSEG0(pa);
	ls = curcpu()->ci_ci.dcache_line_size;

	DPRINTF(DBG_CACHE, ("cache_wb_page(%p, %08lx): va %08lx, ls %d ... ",
	    pm, pa, va, ls));

	for (off = PAGE_SIZE - ls; off >= 0; off -= ls) {
		__asm __volatile("dcbst %0,%1" :: "b"(off), "r"(va));
		__asm __volatile("icbi %0,%1" :: "b"(off), "r"(va));
	}

	DPRINTF(DBG_CACHE, ("done\n"));
}

static __inline void
cache_wbinv_page(pmap_t pm, paddr_t pa)
{
	vaddr_t va;
	int off, ls;

	if (pm && pm->pm_remove_all) {
		DPRINTF(DBG_CACHE, ("cache_wb_page(%p, %08lx): skipping.\n",
		    pm, pa));
		return;
	}

	va = PPC_PA_TO_KSEG0(pa);
	ls = curcpu()->ci_ci.dcache_line_size;

	DPRINTF(DBG_CACHE, ("cache_wb_page(%p, %08lx): va %08lx, ls %d ... ",
	    pm, pa, va, ls));

	for (off = PAGE_SIZE - ls; off >= 0; off -= ls) {
		__asm __volatile("dcbf %0,%1" :: "b"(off), "r"(va));
		__asm __volatile("icbi %0,%1" :: "b"(off), "r"(va));
	}

	DPRINTF(DBG_CACHE, ("done\n"));
}

static __inline void
cache_wb_all(pmap_t pm)
{
	const struct cpu_info *ci;
	vaddr_t va;
	int off, ls;
	register_t dummy;

	if (pm && pm->pm_remove_all) {
		DPRINTF(DBG_CACHE, ("cache_wb_all(%p): skipping.\n", pm));
		return;
	}

	/* Invalidate the instruction cache */
	mtspr(SPR_IC_CST, IC_CST_CMD_INVALL);

	va = PPC_KSEG0_BASE;
	ci = curcpu();
	ls = ci->ci_ci.dcache_line_size;

	DPRINTF(DBG_CACHE, ("cache_wb_all(%p): va %08lx, ls %d ... ",
	    pm, va, ls));

	for (off = ci->ci_ci.dcache_size - ls; off >= 0; off -= ls)
		__asm __volatile("lwzx %0,%1,%2" : "=r"(dummy) :
		    "b"(off), "r"(va));

	DPRINTF(DBG_CACHE, ("done\n"));
}

static inline struct pv_entry *
pa_to_pv(paddr_t pa)
{
	int bank, pg;

	bank = vm_physseg_find(atop(pa), &pg);
	if (bank == -1)
		return NULL;
	return &vm_physmem[bank].pmseg.pvent[pg];
}

static inline char *
pa_to_attr(paddr_t pa)
{
	int bank, pg;

	bank = vm_physseg_find(atop(pa), &pg);
	if (bank == -1)
		return NULL;
	return &vm_physmem[bank].pmseg.attrs[pg];
}

/*
 * Insert PTE into page table.
 */
static inline void
pte_enter(pmap_t pm, vaddr_t va, pt_entry_t pte)
{
	pd_entry_t l1;
 	pt_entry_t *pl2;
 	pt_entry_t oldpte;

	DPRINTF(DBG_PTE_ENTER, ("pte_enter(%p, %08lx, %08x): ", pm, va, pte));

	if (PMAP_L1_IS_NULL(l1 = pm->pm_l1[L1_VA2IDX(va)])) {
		/* Allocate an L2 page table. This will sleep! */
		pl2 = pool_cache_get(&pmap_pt_cache, PR_WAITOK);

		l1 = pm->pm_l1[L1_VA2IDX(va)] =
		    ((pd_entry_t)PPC_KSEG0_TO_PA((vaddr_t)pl2) & L1_L2BA_MASK) |
		    L1_VALID | L1_PS_SMALL;

		pm->pm_stats.resident_count++;
		DPRINTF(DBG_PTE_ENTER, ("new L2 %p, ", pl2));
	}

	pl2 = &((pt_entry_t *)PPC_PA_TO_KSEG0(L1_PDE2PA(l1)))[L2_VA2IDX(va)];

	oldpte = *pl2;
	*pl2 = pte;

	if (oldpte & L2_VALID)
		TLBIE(va);

	DPRINTF(DBG_PTE_ENTER, ("pl2 %p, oldpte %08x\n", pl2, *pl2));

	if (oldpte == 0)
		pm->pm_stats.resident_count++;

	DPRINTF(DBG_PTE_ENTER, ("pte_enter() pte %08x in place\n", pte));
}

/*
 * Get a pointer to a PTE in a page table.
 */
static pt_entry_t *
pte_find(pmap_t pm, vaddr_t va)
{
	pd_entry_t l1;
	pt_entry_t *l2;

	if (PMAP_L1_IS_NULL(l1 = pm->pm_l1[L1_VA2IDX(va)]))
		return (NULL);

	l2 = (pt_entry_t *)PPC_PA_TO_KSEG0(L1_PDE2PA(l1));

	return (&l2[L2_VA2IDX(va)]);
}

/*
 * This is called during initppc, before the system is really initialized.
 */
void
pmap_bootstrap(paddr_t avail_start, psize_t avail_size)
{
	pmap_t pm = pmap_kernel();
	u_int sz;

	avail_start = round_page(avail_start);
	avail_size = trunc_page(avail_size);
	npgs = physmem = btoc(avail_size);

	/*
	 * Announce page-size to the VM-system
	 */
	uvmexp.pagesize = NBPG;
	uvm_setpagesize();

	/*
	 * allow for msgbuf
	 */
	if (msgbuf_paddr == 0) {
		sz = round_page(MSGBUFSIZE);
		npgs -= btoc(sz);
		msgbuf_paddr = avail_start + avail_size - sz;
		avail_size -= sz;
	} else
		sz = 0;

	/*
	 * Give remaining pages to UVM
	 */
	uvm_page_physload(atop(avail_start), atop(avail_start + avail_size),
	    atop(avail_start), atop(avail_start + avail_size),
	    VM_FREELIST_DEFAULT);

	/*
	 * Initialize kernel pmap
	 */
	pm->pm_casid = -1;
	pm->pm_casidgen = 0;
	pm->pm_refs = 1;
	pm->pm_remove_all = 0;
	memset(&pm->pm_stats, 0, sizeof(pm->pm_stats));

	pm->pm_stats.wired_count = pm->pm_stats.resident_count =
	    atop(avail_start + sz);

	pmap_next_casid = 0;
	pmap_casid_generation = 0;
}

static int
pmap_pt_ctor(void *arg, void *v, int flags)
{

	memset(v, 0, PAGE_SIZE);
	return (0);
}

/*
 * Initialize anything else for pmap handling.
 * Called during vm_init().
 */
void
pmap_init(void)
{
	struct pv_entry *pv;
	vsize_t sz;
	vaddr_t addr;
	int i, s;
	int bank;
	char *attr;

	sz = (vsize_t)((sizeof(struct pv_entry) + 1) * npgs);
	sz = round_page(sz);
	addr = uvm_km_alloc(kernel_map, sz, 0, UVM_KMF_WIRED | UVM_KMF_ZERO);
	s = splvm();
	pv = pv_table = (struct pv_entry *)addr;
	for (i = npgs; --i >= 0;)
		pv++->pv_pm = NULL;
	pmap_attrib = (char *)pv;
	memset(pv, 0, npgs);

	pv = pv_table;
	attr = pmap_attrib;
	for (bank = 0; bank < vm_nphysseg; bank++) {
		sz = vm_physmem[bank].end - vm_physmem[bank].start;
		vm_physmem[bank].pmseg.pvent = pv;
		vm_physmem[bank].pmseg.attrs = attr;
		pv += sz;
		attr += sz;
	}

	pmap_initialized = 1;
	splx(s);

	/* Setup a pool for additional pvlist structures */
	pool_init(&pv_pool, sizeof(struct pv_entry), 0, 0, 0, "pv_entry", NULL);

	/* Setup a pool for pmap structures */
	pool_init(&pmap_pool, sizeof(struct pmap), 0, 0, 0, "pmap_t", NULL);

	/* Setup page-table pool cache */
	pool_init(&pmap_pt_pool, PAGE_SIZE, 0, 0, 0, "pt_pool", NULL);
	pool_cache_init(&pmap_pt_cache, &pmap_pt_pool, pmap_pt_ctor, NULL,NULL);
}

/*
 * How much virtual space is available to the kernel?
 */
void
pmap_virtual_space(vaddr_t *start, vaddr_t *end)
{

	/* Skip the first page - we map that early on for device access */
	*start = (vaddr_t) VM_MIN_KERNEL_ADDRESS;
	*end = (vaddr_t) VM_MAX_KERNEL_ADDRESS;
}

/*
 *	vm_page_alloc1:
 *
 *	Allocate and return a memory cell with no associated object.
 */
static struct vm_page *
vm_page_alloc1(void)
{
	struct vm_page *pg = uvm_pagealloc(NULL, 0, NULL, UVM_PGA_USERESERVE);

	if (pg) {
		pg->wire_count = 1;	/* no mappings yet */
		pg->flags &= ~PG_BUSY;	/* never busy */
	}
	return pg;
}

vaddr_t kbreak = VM_MIN_KERNEL_ADDRESS;

vaddr_t
pmap_growkernel(vaddr_t maxkvaddr)
{
	paddr_t pg;
	int s;

	DPRINTF(DBG_GROW_KERNEL, ("pmap_growkernel(%08lx): ", maxkvaddr));

	s = splvm();

	/* Align with the start of a page table */
	kbreak &= ~(L2_PTMAP-1);

	for ( ; kbreak < maxkvaddr; kbreak += L2_PTMAP) {
		if (pte_find(pmap_kernel(), kbreak))
			continue;

		if (uvm.page_init_done) {
			pg = (paddr_t)VM_PAGE_TO_PHYS(vm_page_alloc1());
		} else {
			if (!uvm_page_physget(&pg))
				panic("pmap_growkernel: no memory");
		}

		if (!pg)
			panic("pmap_growkernel: no pages");

		memset((void *)PPC_PA_TO_KSEG0(pg), 0, PAGE_SIZE);

		pmap_kernel()->pm_l1[L1_VA2IDX(kbreak)] =
		    ((pd_entry_t)PPC_KSEG0_TO_PA((vaddr_t)pg) & L1_L2BA_MASK) |
		    L1_VALID | L1_PS_SMALL;
		pmap_kernel()->pm_stats.resident_count++;
	}

	splx(s);

	DPRINTF(DBG_GROW_KERNEL, ("new kbreak %08lx\n", kbreak));

	return (kbreak);
}

/*
 * Create and return a physical map.
 */
pmap_t
pmap_create(void)
{
	pmap_t pm;

	DPRINTF(DBG_CREATE, ("pmap_create(): "));

	pm = pool_get(&pmap_pool, PR_WAITOK);

	pm->pm_casid = -1;
	pm->pm_casidgen = pmap_casid_generation;
	pm->pm_refs = 1;
	pm->pm_remove_all = 0;
	memset(&pm->pm_stats, 0, sizeof(pm->pm_stats));
	pm->pm_stats.resident_count = 1;
	pm->pm_l1 = pool_cache_get(&pmap_pt_cache, PR_WAITOK);

	DPRINTF(DBG_CREATE, ("pm %p, L1 %p\n", pm, pm->pm_l1));

	return pm;
}

/*
 * Add a reference to the given pmap.
 */
void
pmap_reference(pmap_t pm)
{

	pm->pm_refs++;
}

/*
 * Retire the given pmap from service.
 * Should only be called if the map contains no valid mappings.
 */
void
pmap_destroy(pmap_t pm)
{
	int i;

	DPRINTF(DBG_DESTROY, ("pmap_destroy(%p): ", pm));

	if (pm->pm_remove_all)
		pm->pm_remove_all = 0;

	if (--pm->pm_refs > 0) {
		DPRINTF(DBG_DESTROY, ("has other references\n"));
		return;
	}

#ifdef DIAGNOSTIC
	if (pm == pmap_kernel())
		panic("pmap_destroy: freeing kernel/reserved pmap");
#endif

	/* Release any resources held by the given physical map. */
	for (i = 0; i < L1_VA2IDX(PPC_KSEG0_BASE); i++) {
		if (PMAP_L1_IS_NULL(pm->pm_l1[i]))
			continue;

		pool_cache_put(&pmap_pt_cache,
		    (void *)PPC_PA_TO_KSEG0(L1_PDE2PA(pm->pm_l1[i])));

		pm->pm_l1[i] = PMAP_L1_NULL;
	}

	pool_cache_put(&pmap_pt_cache, (void *)PPC_PA_TO_KSEG0(pm->pm_l1));
	pool_put(&pmap_pool, pm);
	TLBSYNC();

	DPRINTF(DBG_DESTROY, ("done\n"));
}

/*
 * Copy the range specified by src_addr/len
 * from the source map to the range dst_addr/len
 * in the destination map.
 *
 * This routine is only advisory and need not do anything.
 */
void
pmap_copy(pmap_t dst_pmap, pmap_t src_pmap, vaddr_t dst_addr,
	  vsize_t len, vaddr_t src_addr)
{

}

/*
 * Require that all active physical maps contain no
 * incorrect entries NOW.
 */
void
pmap_update(pmap_t pm)
{

	DPRINTF(DBG_UPDATE, ("pmap_update(%p): ", pm));

	if (pm->pm_remove_all) {
		DPRINTF(DBG_UPDATE, ("clearing pm_remove_all. "));
		pm->pm_remove_all = 0;
	}

	TLBSYNC();
	__asm __volatile("isync");
	DPRINTF(DBG_UPDATE, ("All done.\n"));
}

/*
 * The vmspace described by this pmap is about to be torn down.
 * Until pmap_update() is called, UVM will only make calls
 * to pmap_remove(). We can make life much simpler by flushing
 * the cache now, and deferring TLB invalidation to pmap_update().
 */
void
pmap_remove_all(pmap_t pm)
{
	const struct cpu_info *ci = curcpu();
	int cs;

	DPRINTF(DBG_REMOVE_ALL, ("pmap_remove_all(%p)\n", pm));

	cs = ci->ci_ci.dcache_size + ci->ci_ci.icache_size;
	if (pm->pm_stats.resident_count >= (cs / PAGE_SIZE)) {
		cache_wb_all(pm);
		pm->pm_remove_all = TRUE;
	}
	DPRINTF(DBG_REMOVE_ALL, ("pmap_remove_all(): all done.\n"));
}

/*
 * Garbage collects the physical map system for
 * pages which are no longer used.
 * Success need not be guaranteed -- that is, there
 * may well be pages which are not referenced, but
 * others may be collected.
 * Called by the pageout daemon when pages are scarce.
 */
void
pmap_collect(pmap_t pm)
{

	/* XXX: Could free up unused L2s */
}

/*
 * Fill the given physical page with zeroes.
 */
void
pmap_zero_page(paddr_t pa)
{

	DPRINTF(DBG_ZERO_PAGE, ("pmap_zero_page(0x%08lx)\n", pa));

	memset((void *)PPC_PA_TO_KSEG0(pa), 0, PAGE_SIZE);
}

/*
 * Copy the given physical source page to its destination.
 */
void
pmap_copy_page(paddr_t src, paddr_t dst)
{

	DPRINTF(DBG_COPY_PAGE, ("pmap_copy_page(src 0x%08lx, dst 0x%08lx)\n",
	    src, dst));

	memcpy((void *)PPC_PA_TO_KSEG0(dst), (void *)PPC_PA_TO_KSEG0(src),
	    PAGE_SIZE);
}

/*
 * This returns whether this is the first mapping of a page.
 */
static inline int
pmap_enter_pv(pmap_t pm, vaddr_t va, paddr_t pa, int wired)
{
	struct pv_entry *pv, *npv = NULL;
	int s;

	if (!pmap_initialized)
		return 0;

	s = splvm();

	pv = pa_to_pv(pa);

#ifdef DIAGNOSTIC	/* XXX */
	for (npv = pv; npv; npv = npv->pv_next)
		if (PV_VA(npv) == va && npv->pv_pm == pm) {
			printf("Duplicate pv: va %lx pm %p\n", va, pm);
#ifdef DDB
			Debugger();
#endif
			return (1);
		}
#endif /* DIAGNOSTIC */

	if (!pv->pv_pm) {
		/*
		 * No entries yet, use header as the first entry.
		 */
		pv->pv_va = va;
		pv->pv_pm = pm;
		pv->pv_next = NULL;
	} else {
		/*
		 * There is at least one other VA mapping this page.
		 * Place this entry after the header.
		 */
		npv = pool_get(&pv_pool, PR_WAITOK);
		if (!npv) return (0);
		npv->pv_va = va;
		npv->pv_pm = pm;
		npv->pv_next = pv->pv_next;
		pv->pv_next = npv;
	}
	splx(s);
	return (1);
}

static void
pmap_remove_pv(pmap_t pm, vaddr_t va, paddr_t pa)
{
	struct pv_entry *pv, *npv;

	/*
	 * Remove from the PV table.
	 */
	pv = pa_to_pv(pa);
	if (!pv) return;

	/*
	 * If it is the first entry on the list, it is actually
	 * in the header and we must copy the following entry up
	 * to the header.  Otherwise we must search the list for
	 * the entry.  In either case we free the now unused entry.
	 */
	if (pm == pv->pv_pm && PV_CMPVA(va, pv)) {
		if ((npv = pv->pv_next)) {
			*pv = *npv;
			pool_put(&pv_pool, npv);
		} else
			pv->pv_pm = NULL;
	} else {
		for (; (npv = pv->pv_next) != NULL; pv = npv)
			if (pm == npv->pv_pm && PV_CMPVA(va, npv))
				break;
		if (npv) {
			pv->pv_next = npv->pv_next;
			pool_put(&pv_pool, npv);
		}
	}
}

/*
 * Insert physical page at pa into the given pmap at virtual address va.
 */
int
pmap_enter(pmap_t pm, vaddr_t va, paddr_t pa, vm_prot_t prot, int flags)
{
	pt_entry_t *ptep, pte, opte;
	char *pattr, *poattr;
	int s;

	DPRINTF(DBG_ENTER, ("pmap_enter(%p, %08lx, %08lx, %04x, %04x) %s\n",
	    pm, va, pa, prot, flags, pmap_initialized ? "" : "pre-pmap_init!"));

	KDASSERT((flags & PMAP_WIRED) == 0 || (flags & VM_PROT_ALL) != 0);
	KDASSERT(((va | pa) & PGOFSET) == 0);

	pte = pa | L2_SZ_4KB;

	if (pm == pmap_kernel()) {
		KDASSERT(va >= PPC_KSEG1_BASE);
		pte |= (prot & VM_PROT_WRITE) ? L2_PP_K_RDWR : L2_PP_K_RO;
		pte |= L2_SHARED;
	} else {
		KDASSERT(va < PPC_KSEG0_BASE);
		pte |= (prot & VM_PROT_WRITE) ? L2_PP_U_RDWR : L2_PP_U_RO;
	}

	pattr = pmap_initialized ? pa_to_attr(pa) : NULL;

	if ((ptep = pte_find(pm, va)) != NULL)
		opte = *ptep;
	else
		opte = 0;

	if (opte && pmap_initialized) {
		/*
		 * There is already a mapping at this address.
		 * If this physical address is different, lookup its
		 * attributes
		 */
		if (L2_PTE2PA(opte) != pa)
			poattr = pa_to_attr(L2_PTE2PA(opte));
		else
			poattr = pattr;
	} else
		poattr = NULL;

	if (pattr) {
		/*
		 * This is to be a managed mapping
		 */
		DPRINTF(DBG_ENTER, ("pmap_enter(): managed mapping"));
		if ((flags & (VM_PROT_READ | VM_PROT_EXECUTE)) ||
		    (*pattr & PMAP_REF)) {
			/*
			 * - The access type indicates that we don't need
			 *   to do referenced emulation.
			 * OR
			 * - The physical page has already been referenced
			 *   so no need to re-do referenced emulation here.
			 */
			pte |= L2_VALID;
			*pattr |= PMAP_REF;
			DPRINTF(DBG_ENTER, (", L2_VALID"));
		}

		if ((flags & VM_PROT_WRITE) || (*pattr & PMAP_MOD)) {
			/*
			 * This is a writable mapping, and the page's mod
			 * state indicates it has already been modified.
			 * Make it writable from the outset.
			 */
			pte |= L2_CHANGE;
			*pattr |= PMAP_MOD;
			DPRINTF(DBG_ENTER, (", L2_CHANGE"));
		}

		if (pattr == poattr) {
			struct pv_entry *pv;

			/*
			 * We're changing the attributes of an existing
			 * mapping.
			 * If we're doing rw-ro, then write-back the cache.
			 */
			DPRINTF(DBG_ENTER, (", existing"));
			if ((prot & VM_PROT_WRITE) == 0 &&
			    (pte & L2_PP_MASK) != (opte & L2_PP_MASK))
				cache_wb_page(pm, pa);

			/*
			 * Update wiring
			 *
			 * XXX: Todo
			 * Track wiring in an unused PTE bit, to enable
			 * pmap_collect() to be fleshed out.
			 */
			s = splvm();
			for (pv = pa_to_pv(pa); pv != NULL; pv = pv->pv_next) {
				if (pv->pv_pm == pm && PV_CMPVA(va, pv))
					break;
			}

			if (pv) {
				int cnt;

				if (flags & PMAP_WIRED) {
					cnt = PV_IS_WIRED(pv) ? 0 : 1;
					PV_WIRE(pv);
				} else {
					cnt = PV_IS_WIRED(pv) ? -1 : 0;
					PV_UNWIRE(pv);
				}

				pm->pm_stats.wired_count += cnt;
			}
			splx(s);
		} else {
			/*
			 * New mapping, or changing the backing page of
			 * an existing mapping.
			 */
			DPRINTF(DBG_ENTER, (", new mapping"));

			if (prot & VM_PROT_EXECUTE) {
				/*
				 * Unconditionally write-back the data cache
				 * and sync the instruction cache if this
				 * is an executable mapping.
				 */
				cache_wb_page(NULL, pa);
			}

			if (poattr) {
				/*
				 * Replacing an existing mapping with a new one.
				 * It is part of our managed memory so we
				 * must remove it from the PV list.
				 */
				DPRINTF(DBG_ENTER, (", [change backing page]"));
				pmap_remove_pv(pm, va, L2_PTE2PA(opte));

				/*
				 * If the old mapping was writable, make sure
				 * the dcache gets written back.
				 * XXX: This may be overkill.
				 */
				if ((opte & L2_CHANGE) &&
				    (prot & VM_PROT_EXECUTE) == 0)
					cache_wb_page(pm, pa);
			}

			if (!pmap_enter_pv(pm, va, pa, flags & PMAP_WIRED)) {
				if ((flags & PMAP_CANFAIL) == 0)
					panic("pmap_enter: no pv " "entries");
				return (ENOMEM);
			}
		}
		DPRINTF(DBG_ENTER, ("\n"));
	} else {
		/*
		 * We're mapping an unmanaged page.
		 * These are always readable, and possibly writable, from
		 * the get go as we don't need to track ref/mod status.
		 */
		DPRINTF(DBG_ENTER, ("pmap_enter(): unmanaged mapping"));
		pte |= L2_CHANGE | L2_VALID;

		if (poattr) {
			/*
			 * Looks like there's an existing 'managed' mapping
			 * at this address.
			 */
			DPRINTF(DBG_ENTER, (", existing managed"));
			pmap_remove_pv(pm, va, pa);

			/*
			 * If the old mapping was writable, flush
			 * the page from the cache.
			 */
			if (opte & L2_CHANGE)
				cache_wb_page(pm, pa);
		}
		DPRINTF(DBG_ENTER, ("\n"));
	}

	/* I/O mapping? */
	if (flags & PME_NOCACHE)
		pte |= L2_CI;

	s = splvm();

	/* Insert page into page table. */
	pte_enter(pm, va, pte);

	splx(s);

	DPRINTF(DBG_ENTER, ("pmap_enter(): all done\n"));

	return 0;
}

void
pmap_unwire(pmap_t pm, vaddr_t va)
{
	struct pv_entry *pv, *npv;
	paddr_t pa;
	int s = splvm();

	if (pm == NULL || !pmap_extract(pm, va, &pa))
		return;

	va |= PV_WIRED;

	if ((pv = pa_to_pv(pa)) == NULL)
		return;

	/*
	 * If it is the first entry on the list, it is actually
	 * in the header and we must copy the following entry up
	 * to the header.  Otherwise we must search the list for
	 * the entry.  In either case we free the now unused entry.
	 */
	for (npv = pv; (npv = pv->pv_next) != NULL; pv = npv) {
		if (pm == npv->pv_pm && va == npv->pv_va) {
			PV_UNWIRE(npv);
			pm->pm_stats.wired_count--;
			break;
		}
	}
	splx(s);
}

void
pmap_kenter_pa(vaddr_t va, paddr_t pa, vm_prot_t prot)
{
	int s;
	pt_entry_t pte;
	pmap_t pm = pmap_kernel();

	DPRINTF(DBG_KENTER, ("pmap_kenter_pa(%08lx, %08lx, %04x)\n",
	    va, pa, prot));

	/*
	 * Generate PTE.
	 */
	pte = (pt_entry_t)(pa & L2_RPN_MASK) | L2_SZ_4KB | L2_SHARED | L2_VALID;

	/* I/O mapping? */
	if (prot & PME_NOCACHE)
		pte |= L2_CI;

	if (prot & VM_PROT_WRITE)
		pte |= L2_PP_K_RDWR | L2_CHANGE;
	else
		pte |= L2_PP_K_RO;

	s = splvm();

	/* Insert page into page table. */
	pte_enter(pm, va, pte);
	splx(s);

	DPRINTF(DBG_KENTER, ("pmap_kenter_pa(): all done\n"));
}

void
pmap_kremove(vaddr_t va, vsize_t len)
{
	const struct cpu_info *ci = curcpu();
	pt_entry_t *pl2, l2;
	pmap_t pm;

	DPRINTF(DBG_KREMOVE, ("pmap_kremove(%08lx, %08lx)\n", va, len));

	pm = pmap_kernel();
	if (pm->pm_remove_all == 0 && (len / PAGE_SIZE) >=
	    (ci->ci_ci.dcache_size + ci->ci_ci.icache_size)) {
		cache_wb_all(pm);
		pm->pm_remove_all = 1;
	}

	for (; len > 0; len -= PAGE_SIZE, va += PAGE_SIZE) {
		if ((pl2 = pte_find(pm, va)) == NULL || (l2 = *pl2) == 0)
			continue;

		DPRINTF(DBG_KREMOVE, ("pmap_kremove(): va %08lx, pte %08x\n",
		    va, l2));

		if (l2 & L2_CHANGE)
			cache_wb_page(pm, L2_PTE2PA(l2));
		*pl2 = 0;
		TLBIE(va);
		pm->pm_stats.resident_count--;
	}

	DPRINTF(DBG_KREMOVE, ("pmap_kremove(): all done\n"));
}

/*
 * Remove the given range of mapping entries.
 */
void
pmap_remove(pmap_t pm, vaddr_t va, vaddr_t endva)
{
	const struct cpu_info *ci = curcpu();
	pt_entry_t *pl2, l2;
	paddr_t pa;
	int s;

	DPRINTF(DBG_REMOVE, ("pmap_remove(%p, %08lx, %08lx)\n", pm, va, endva));

	if (pm->pm_remove_all == 0 && ((endva - va) / PAGE_SIZE) >=
	    (ci->ci_ci.dcache_size + ci->ci_ci.icache_size)) {
		cache_wb_all(pm);
		pm->pm_remove_all = 1;
	}

	s = splvm();
	for (; va < endva; va += PAGE_SIZE) {
		if ((pl2 = pte_find(pm, va)) == NULL || (l2 = *pl2) == 0)
			continue;

		DPRINTF(DBG_REMOVE, ("pmap_remove(): va %08lx, pte %08x\n",
		    va, l2));

		pa = L2_PTE2PA(l2);

		if (l2 & L2_CHANGE)
			cache_wb_page(pm, pa);
		pmap_remove_pv(pm, va, pa);
		*pl2 = 0;

		/*
		 * XXXSCW:
		 * Maintain a per-L2 ref count so we can free up L2s as well
		 */

		if (l2 & L2_VALID)
			TLBIE(va);
		pm->pm_stats.resident_count--;
	}
	splx(s);

	DPRINTF(DBG_REMOVE, ("pmap_remove(): all done\n"));
}

/*
 * Get the physical page address for the given pmap/virtual address.
 */
boolean_t
pmap_extract(pmap_t pm, vaddr_t va, paddr_t *pap)
{
	pd_entry_t l1;
	pt_entry_t *pl2, l2;
	paddr_t pa;

	DPRINTF(DBG_EXTRACT, ("pmap_extract(%p, %08lx): ", pm, va));

	/*
	 * KSEG0 is easy.
	 */
	if (pm == pmap_kernel() &&
	    va < PPC_KSEG1_BASE && va >= PPC_KSEG0_BASE) {
		if (pap) {
			*pap = PPC_KSEG0_TO_PA(va);
			DPRINTF(DBG_EXTRACT, ("KSEG0 - %08lx\n", *pap));
		}
		return (1);
	}

	if (PMAP_L1_IS_NULL(l1 = pm->pm_l1[L1_VA2IDX(va)])) {
		DPRINTF(DBG_EXTRACT, ("L1 slot is empty\n"));
		return (0);
	}

	pl2 = (pt_entry_t *)PPC_PA_TO_KSEG0(L1_PDE2PA(l1));

	switch (l1 & L1_PS_MASK) {
	case L1_PS_SMALL:
		/* Standard 4KB page size */
		l2 = pl2[L2_VA2IDX(va)];
		pa = L2_PTE2PA(l2);
		pa |= (va & PGOFSET);
		break;

	case L1_PS_8MB:
		l2 = pl2[0];
		pa = (paddr_t) (l2 & 0xff800000u);
		pa |= (va & 0x007fffffu);
		break;

	case L1_PS_512KB:
	default:		/* XXX: Can't happen! */
		l2 = 0;
		pa = 0;
		break;
	}

	DPRINTF(DBG_EXTRACT, ("rv %d, pa %08lx\n", (l2 != 0), pa));

	if (pap)
		*pap = pa;
	return (l2 != 0);
}

/*
 * Lower the protection on the specified range of this pmap.
 *
 * There are only two cases: either the protection is going to 0,
 * or it is going to read-only.
 */
static void
pmap_protect2(pmap_t pm, vaddr_t sva, vaddr_t eva, vm_prot_t prot, int wb)
{
	pt_entry_t *pl2, l2, nl2, pte;

	if ((prot & VM_PROT_READ) == 0) {
		pmap_remove(pm, sva, eva);
		return;
	}

	if (prot & VM_PROT_WRITE)
		return;

	if (sva < PPC_KSEG0_BASE) {
		KDASSERT(eva < PPC_KSEG0_BASE);
		pte = L2_PP_U_RO;
	} else
		pte = L2_PP_K_RO;

	for (; sva < eva; sva += PAGE_SIZE) {
		if ((pl2 = pte_find(pm, sva)) != NULL && (l2 = *pl2) != 0) {
			nl2 = (l2 & ~L2_PP_MASK) | pte;
			if (nl2 == l2)
				continue;

			if (wb && (l2 & L2_CHANGE))
				cache_wb_page(pm, L2_PTE2PA(l2));

			*pl2 = nl2;

			if (l2 & L2_VALID)
				TLBIE(sva);
		}
	}
}

void
pmap_protect(pmap_t pm, vaddr_t sva, vaddr_t eva, vm_prot_t prot)
{
	int s;

	DPRINTF(DBG_PROTECT, ("pmap_protect(%p, %08lx, %08lx, %04x)\n",
	    pm, sva, eva, prot));

	s = splvm();
	pmap_protect2(pm, sva, eva, prot, 1);
	splx(s);

	DPRINTF(DBG_PROTECT, ("pmap_protect(): all done\n"));
}

/*
 * Lower the protection on the specified physical page.
 *
 * There are only two cases: either the protection is going to 0,
 * or it is going to read-only.
 */
void
pmap_page_protect(struct vm_page *pg, vm_prot_t prot)
{
	paddr_t pa = VM_PAGE_TO_PHYS(pg);
	vaddr_t va;
	struct pv_entry *pvh, *pv, *npv;
	pmap_t pm;
	int s;

	DPRINTF(DBG_PAGE_PROTECT, ("pmap_page_protect(%08lx, %04x)", pa, prot));

	pvh = pa_to_pv(pa);
	if (pvh == NULL) {
		DPRINTF(DBG_PAGE_PROTECT, (": not a managed page!\n"));
		return;
	}

	DPRINTF(DBG_PAGE_PROTECT, ("\n"));

	s = splvm();

	if ((prot & VM_PROT_READ) != 0)
		cache_wb_page(NULL, pa);

	/* Handle extra pvs which may be deleted in the operation */
	for (pv = pvh->pv_next; pv; pv = npv) {
		npv = pv->pv_next;
		pm = pv->pv_pm;
		va = PV_VA(pv);
		DPRINTF(DBG_PAGE_PROTECT,
		    ("pmap_page_protect(): pm %p, va %08lx\n", pm, va));
		pmap_protect2(pm, va, va + PAGE_SIZE, prot, 0);
	}
	/* Now check the head pv */
	if (pvh->pv_pm) {
		pv = pvh;
		pm = pv->pv_pm;
		va = PV_VA(pv);
		DPRINTF(DBG_PAGE_PROTECT,
		    ("pmap_page_protect(): pm %p, va %08lx\n", pm, va));
		pmap_protect2(pm, va, va + PAGE_SIZE, prot, 0);
	}

	splx(s);

	DPRINTF(DBG_PAGE_PROTECT, ("pmap_page_protect(): all done"));
}

boolean_t
pmap_check_attr(struct vm_page *pg, u_int mask)
{
	paddr_t pa = VM_PAGE_TO_PHYS(pg);
	struct pv_entry *pv;
	char *pattr, attr;
	pt_entry_t *pl2, l2, l2bits;
	vaddr_t va;
	int s, rv;

	DPRINTF(DBG_REFMOD, ("pmap_check_attr(%08lx, %s%s): ", pa,
	    (mask & PMAP_REF) ? "REF" : "MOD",
	    (mask & PMAP_CLEAR) ? " | CLR" : ""));

	pattr = pa_to_attr(pa);
	if (pattr == NULL) {
		DPRINTF(DBG_REFMOD, ("not a managed page!\n"));
		return FALSE;
	}

	attr = *pattr;
	rv = ((attr & mask) != 0);

	if ((mask & PMAP_CLEAR) && rv) {
		DPRINTF(DBG_REFMOD, ("clearing bit\n"));
		if (mask & PMAP_REF) {
			l2bits = L2_VALID;
			cache_wbinv_page(NULL, pa);
		} else {
			l2bits = L2_CHANGE;
			cache_wb_page(NULL, pa);
		}

		s = splvm();
		pv = pa_to_pv(pa);
		if (pv->pv_pm) {
			for ( ; pv != NULL; pv = pv->pv_next) {
				va = PV_VA(pv);
				pl2 = pte_find(pv->pv_pm, va);
				KASSERT(pl2 != NULL);
				KASSERT(*pl2 != 0);

				l2 = *pl2;

				DPRINTF(DBG_REFMOD,
				    ("pmap_check_attr(): %08x -> %08x\n",
				    l2, l2 & ~l2bits));

				*pl2 = l2 & ~l2bits;
				if (l2 & L2_VALID)
					TLBIE(va);
			}
		}
		*pattr &= ~mask;
		splx(s);
		DPRINTF(DBG_REFMOD, ("pmap_check_attr(): all done: "));
	}

	DPRINTF(DBG_REFMOD, ("rv = %d\n", rv));

	return rv;
}

static inline void
pmap_casid_alloc(pmap_t pm)
{

	KDASSERT(pm != pmap_kernel());

	if (pm->pm_casid < 0 || pm->pm_casidgen != pmap_casid_generation) {
		if ((pmap_next_casid & M_CASID_MASK) != pmap_next_casid) {
			TLBIA();
			TLBSYNC();
			pmap_casid_generation++;
			pmap_next_casid = 0;
			DPRINTF(DBG_ACTIVATE, ("pmap_casid_alloc(): gen++\n"));
		}
		pm->pm_casid = pmap_next_casid++;
		pm->pm_casidgen = pmap_casid_generation;
	}
}

/*
 * Activate the address space for the specified process.  If the process
 * is the current process, load the new MMU context.
 */
void
pmap_activate(struct lwp *l)
{
	pmap_t pm = l->l_proc->p_vmspace->vm_map.pmap;

	l->l_addr->u_pcb.pcb_pm = pm;

	DPRINTF(DBG_ACTIVATE, ("pmap_activate(%p): pm %p\n", l, pm));

	if (pm != pmap_kernel())
		pmap_casid_alloc(pm);
	if (l == curlwp) {
		mtspr(SPR_M_CASID, pm->pm_casid);
		mtspr(SPR_M_TWB, PPC_KSEG0_TO_PA((vaddr_t)pm->pm_l1));
		DPRINTF(DBG_ACTIVATE, ("pmap_activate(): casid %d, l1 %p\n",
		    pm->pm_casid, pm->pm_l1));
	}

	DPRINTF(DBG_ACTIVATE, ("pmap_activate(): all done\n"));
}

/*
 * Deactivate the specified process's address space.
 */
void
pmap_deactivate(struct lwp *l)
{

	DPRINTF(DBG_ACTIVATE, ("pmap_dactivate(%p)\n", l));
}

/*
 * Synchronize caches corresponding to [addr, addr+len) in p.
 */
void
pmap_procwr(struct proc *p, vaddr_t va, size_t len)
{
	const struct cpu_info *ci = curcpu();
	pmap_t pm = p->p_vmspace->vm_map.pmap;
	vaddr_t cva;
	paddr_t pa;
	u_int cl;

	if (len == 0)
		return;

	cl = min(ci->ci_ci.dcache_line_size, ci->ci_ci.icache_line_size);

	va &= ~(cl - 1);

	if (pmap_extract(pm, va, &pa) == 0)
		return;

	for (len = roundup(len, cl); len; len -= cl) {
		cva = PPC_PA_TO_KSEG0(pa);
		__asm __volatile("dcbst 0,%0" :: "r"(cva));
		__asm __volatile("icbi 0,%0" :: "r"(cva));

		pa += cl;
		va += cl;
		if ((pa & PGOFSET) == 0 && pmap_extract(pm, va, &pa) == 0)
			pa = 0;
	}
}

/*
 * This routine gets first dibs on any write fault in order to track
 * page modification.
 * If it returns non-zero, the fault does not need to be passed to UVM.
 */
int
pmap_write_fault_fixup(pmap_t pm, vaddr_t va, int usermode)
{
	pt_entry_t *pl2, l2;
	char *attr;

	DPRINTF(DBG_FIXUP, ("pmap_write_fault_fixup(%p, %08lx, %d): ",
	    pm, va, usermode));

	/*
	 * Does a mapping exist at this address?
	 */
	if ((pl2 = pte_find(pm, va)) == NULL) {
		DPRINTF(DBG_FIXUP, ("no L1 entry.\n"));
		return (0);
	}

	if ((l2 = *pl2) == 0) {
		DPRINTF(DBG_FIXUP, ("no L2 entry.\n"));
		return (0);
	}

	/*
	 * If the page is valid and L2_CHANGE is already set, then the
	 * page fault cannot be as a result of modify tracking.
	 */
	if ((l2 & (L2_CHANGE|L2_VALID)) == (L2_CHANGE|L2_VALID)) {
		DPRINTF(DBG_FIXUP, ("L2_CHANGE|L2_VALID already set.\n"));
		return (0);
	}

	/*
	 * Is the page actually writable?
	 */
	switch (l2 & L2_PP_MASK) {
	case L2_PP_K_RO:
	case L2_PP_U_RO:
		/*
		 * Read-only page. Nothing more we can do.
		 */
		DPRINTF(DBG_FIXUP, ("read-only page.\n"));
		return (0);

	case L2_PP_K_RDWR:
		/*
		 * Kernel writeable page. Bail if the fault came from usermode.
		 */
		if (usermode) {
			DPRINTF(DBG_FIXUP, ("user write to kernel address.\n"));
			return (0);
		}
		break;

	case L2_PP_U_RDWR:
		/*
		 * User writable page. Anyone can write to it.
		 */
		break;

	default:
		panic("pmap_write_fault_fixup: bogus L2 protection bits: "
		    "pm %p, va %08lx, l2 %08x", pm, va, l2);
		break;
	}

	/*
	 * We have a winner. Simply set L2_CHANGE in the page table
	 * and flush the old entry from the TLB if it was previously valid.
	 */
	DPRINTF(DBG_FIXUP, ("%08x -> %08x, ", l2, l2 | L2_CHANGE | L2_VALID));
	*pl2 = l2 | L2_CHANGE | L2_VALID;
	if (l2 & L2_VALID) {
		DPRINTF(DBG_FIXUP, ("sync TLB/cache, "));
		cache_wbinv_page(pm, L2_PTE2PA(l2));
		TLBIE(va);
		TLBSYNC();
	}

	/*
	 * Update page referenced/modified attributes for the page.
	 */
	if ((attr = pa_to_attr(L2_PTE2PA(l2))) == NULL) {
		/* XXX: Should we panic here? */
		DPRINTF(DBG_FIXUP, ("unmanaged page!\n"));
		return (1);
	}

	DPRINTF(DBG_FIXUP, ("attr %02x -> %02x\n", *attr, PMAP_REF | PMAP_MOD));
	*attr |= PMAP_REF | PMAP_MOD;

	return (1);
}

/*
 * This routine gets first dibs on any read fault in order to track
 * page referenced.
 * If it returns non-zero, the fault does not need to be passed to UVM.
 */
int
pmap_read_fault_fixup(pmap_t pm, vaddr_t va, int usermode)
{
	pt_entry_t *pl2, l2;
	char *attr;

	DPRINTF(DBG_FIXUP, ("pmap_read_fault_fixup(%p, %08lx, %d): ",
	    pm, va, usermode));

	/*
	 * Does a mapping exist at this address?
	 */
	if ((pl2 = pte_find(pm, va)) == NULL) {
		DPRINTF(DBG_FIXUP, ("no L1 entry.\n"));
		return (0);
	}

	if ((l2 = *pl2) == 0) {
		DPRINTF(DBG_FIXUP, ("no L2 entry.\n"));
		return (0);
	}

	/*
	 * If the mapping is marked as valid, then the page fault
	 * cannot be as a result of reference tracking.
	 */
	if ((l2 & L2_VALID) != 0) {
		DPRINTF(DBG_FIXUP, ("mapping already valid: %08x\n", l2));
		return (0);
	}

	/*
	 * Is the page actually readable?
	 */
	if (usermode) {
		switch (l2 & L2_PP_MASK) {
		case L2_PP_K_RO:
		case L2_PP_K_RDWR:
			DPRINTF(DBG_FIXUP, ("user read of kernel address.\n"));
			return (0);

		default:
			break;
		}
	}

	/*
	 * We have a winner. Simply set L2_VALID in the page table.
	 * We don't need to flush the old entry from the TLB, since
	 * it has never been valid until now.
	 */
	DPRINTF(DBG_FIXUP, ("%08x -> %08x, ", l2, l2 | L2_VALID));
	*pl2 = l2 | L2_VALID;

	/*
	 * Update page-referenced attribute for the page.
	 */
	if ((attr = pa_to_attr(L2_PTE2PA(l2))) == NULL) {
		/* XXX: Should we panic here? */
		DPRINTF(DBG_FIXUP, ("unmanaged page!\n"));
		return (1);
	}

	DPRINTF(DBG_FIXUP, ("attr %02x -> %02x\n", *attr, PMAP_REF));
	*attr |= PMAP_REF;

	return (1);
}
